import asyncio
import time
import pathlib

from .context import Context
from .file import File

__all__ = ['TaskExists', 'TaskID', 'Task']

class TaskExists(Exception):
    def __init__(self, id, task):
        super().__init__(f'Task with id {id} already exists')
        self.id = id
        self.task = task

class TaskID(tuple):
    def __new__(cls, *args):
        assert args

        if len(args) == 1 and isinstance(args[0], tuple):
            args, = args

        for i, e in enumerate(args):
            for type in [str, pathlib.Path, int]:
                if isinstance(e, type):
                    break
            else:
                raise TypeError(f'TaskID element {i} is unsupported: {e}')
        
        return super().__new__(cls, (str(e) for e in args))

    @property
    def mangled(self):
        return ';'.join(self)

    @property
    def str(self):
        return ' '.join(self)

class Task:
    ctx: Context
    id: TaskID

    def __new__(cls, ctx, id):
        id = TaskID(id)
        if id in ctx.tasks:
            raise TaskExists(id, ctx.tasks[id])
        ctx.tasks[id] = self = super().__new__(cls)
        self.ctx = ctx
        self.id = id
        self.dependencies = []
        self.lock = asyncio.Lock()
        self.done = False
        self.result = None

        self._input_files = []
        self._output_files = []
        self._start = None
        self._end = None
        return self

    def add_input_files(self, *files):
        files = [File(self.ctx, path) for path in files]
        self._input_files.extend(files)

    def add_output_files(self, *files):
        files = [File(self.ctx, path) for path in files]
        self._output_files.extend(files)
        for file in files:
            assert file.generator_task is None
            file.generator_task = self

    def input_metadata(self):
        return {}

    def dynamic_deps(self):
        return []

    async def run(self):
        raise NotImplementedError()

    async def post_run(self):
        pass

    def _uptodate(self):
        # Uncached tasks are not up to date.
        if not self.id.mangled in self.ctx.cache:
            return False
        cache = self.ctx.cache[self.id.mangled]

        # Check if input metadata changed.
        if cache.get('input_metadata') != self.input_metadata():
            return False

        # Check if any files doesn't match their fingerprints.
        for path, fingerprint in cache.get('file_fingerprints', {}).items():
            if not fingerprint.check(path):
                return False

        ## TODO: This condition should be updated.
        #if not self._input_files or not self._output_files:
        #    return False
        
        for f in self._input_files:
            assert f.path.exists(), f'Required file {f.path} for task {self.id} does not exist.'
        
        for f in self._output_files:
            if not f.path.exists():
                return False
        
        #if max(f.path.stat().st_mtime for f in self._input_files) > min(f.path.stat().st_mtime for f in self._output_files):
        #    return False

        return True

    def _save_cache(self):
        self.ctx.cache[self.id.mangled] = {
            'input_metadata': self.input_metadata(),
            'file_fingerprints': {f.path: f.get_fingerprint() for f in [*self._input_files, *self._output_files]},
            'result': self.result,
        }

    async def _run(self):
        async with self.lock:
            if self.done:
                return self.result

            await async_run(self.dependencies + self._input_files)

            await async_run(self.dynamic_deps())

            await async_run(self._input_files)

            async with self.ctx.task_semaphore:
                self._start = time.monotonic()
                if self._uptodate():
                    self.result = self.ctx.cache[self.id.mangled]['result']
                else:
                    self.result = await self.run()
                    self._save_cache()
                await self.post_run()
                self._end = time.monotonic()

            self.done = True

async def async_run(tasks):
    async with asyncio.TaskGroup() as tg:
        for task in tasks:
            tg.create_task(task._run())



import pathlib

from .context import Context, get_global_context

__all__ = ['Env']

class Env:
    ctx: Context
    build_dir: pathlib.Path

    def __init__(self, *, ctx = None, build_dir = None):
        self.ctx = ctx or get_global_context()
        self.build_dir = pathlib.Path(build_dir or 'build/')



import sys
import asyncio

class Handler:
    def __init__(self, cmi_dir, reader, writer):
        self.cmi_dir = cmi_dir
        self.reader = reader
        self.writer = writer

    async def read(self):
        while line := await self.reader.readline():
            #print(f'< {line!r}', file=sys.stderr)
            yield tuple(line.decode('utf-8').strip().split())

    def write(self, line):
        line = line + '\n'
        self.writer.write(line.encode('utf-8'))
        #print(f'> {line!r}', file=sys.stderr)

    def handle(self, command):
        match command:
            case ('HELLO', '1', *_):
                return 'HELLO 1 erect-modmap'

            case ('MODULE-REPO',):
                return f'PATHNAME {self.cmi_dir}'

            case ('MODULE-EXPORT', module, *_):
                return f'PATHNAME {module}.gcm'

            case ('MODULE-IMPORT', module, *_):
                return f'PATHNAME {module}.gcm'

            case ('MODULE-COMPILED', module):
                return 'OK'

            case ('INCLUDE-TRANSLATE', path):
                return 'BOOL FALSE'

            case _:
                return 'ERROR'

    async def run(self):
        command_queue = []
        async for line in self.read():
            command_queue.append(line)
            if command_queue[-1][-1] == ';':
                *command_queue[-1], _ = command_queue[-1]
                continue

            while len(command_queue) > 1:
                self.write(self.handle(command_queue.pop(0)) + ' ;')
            self.write(self.handle(command_queue.pop(0)))
            await self.writer.drain()
        self.writer.close()

async def start_tcp(cmi_dir):
    async def handle_client(reader, writer):
        m = Handler(cmi_dir, reader, writer)
        await m.run()

    return await asyncio.start_server(handle_client, host = '::1', port = None)

class ModuleMapper:
    def __init__(self, cmi_dir):
        self.port = None
        self.cmi_dir = cmi_dir

    async def start(self):
        server = await start_tcp(self.cmi_dir)
        self.port = server.sockets[0].getsockname()[1]

    @property
    def gcc_arg(self):
        assert self.port is not None
        return f'-fmodule-mapper=localhost:{self.port}'

async def connect_stdin_stdout():
    loop = asyncio.get_event_loop()
    reader = asyncio.StreamReader()
    protocol = asyncio.StreamReaderProtocol(reader)
    await loop.connect_read_pipe(lambda: protocol, sys.stdin)
    w_transport, w_protocol = await loop.connect_write_pipe(asyncio.streams.FlowControlMixin, sys.stdout)
    writer = asyncio.StreamWriter(w_transport, w_protocol, reader, loop)
    return reader, writer

async def async_main(cmi_dir):
    reader, writer = await connect_stdin_stdout()

    m = Handler(cmi_dir, reader, writer)
    await m.run()

def main():
    if len(sys.argv) > 1:
        cmi_dir = sys.argv[1]
    else:
        cmi_dir = 'build/cmi/'

    asyncio.run(async_main(cmi_dir))



import asyncio
import shelve

__all__ = ['Context']

_global_context = None

class Context:
    def __init__(self, *,
        max_concurrent_tasks = None,
        cache_file = None,
    ):
        self.tasks = {}
        self.files = {}
        self._start_coros = []

        self.task_semaphore = asyncio.Semaphore(max_concurrent_tasks or 1)

        if cache_file is False:
            self.cache = {}
        else:
            self.cache = shelve.open(cache_file or '.erect')

    def __enter__(self):
        global _global_context
        assert _global_context is None, 'Global context already exists.'

        _global_context = self
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        global _global_context
        assert _global_context is self, 'Global context is not self.'

        _global_context = None

    def start_async(self, coro):
        self._start_coros.append(coro)

    async def run(self, tasks):
        for coro in self._start_coros:
            await coro

        async with asyncio.TaskGroup() as tg:
            for task in tasks:
                tg.create_task(task._run())

def get_global_context():
    assert _global_context is not None, 'Global context is not set.'

    return _global_context



from ..core.task import Task
from ..core.env import Env

import pathlib
import jinja2

__all__ = ['Jinja2']

class Env(Env):
    pass

loader = jinja2.FileSystemLoader('.')

jinja2_env = jinja2.Environment(
    loader = loader,
    trim_blocks = True,
    lstrip_blocks = True,
)

jinja2_env.filters['hex'] = lambda value: '%#x' % value
jinja2_env.filters['size_prefix'] = lambda value: '%d%s' % next((value / 1024**i, c) for i, c in [(2, 'M'), (1, 'k'), (0, '')] if value % 1024**i == 0)

class Jinja2(Task):
    def __new__(cls, env, target, source, **kwargs):
        self = super().__new__(cls, env.ctx, ('jinja2', env.build_dir, target))
        self.env = env
        self.source = pathlib.Path(source)
        self.target = env.build_dir / 'generated' / target
        self.data = kwargs
        self.add_input_files(self.source)
        self.add_output_files(self.target)
        return self

    def input_metadata(self):
        return super().input_metadata() | {
            'source': self.source,
            'data': self.data,
        }

    async def run(self):
        print(self.id.str)

        template = jinja2_env.get_template(str(self.source))
        output = template.render(**self.data) + '\n'


        # Ensure output directory exists.
        self.target.parent.mkdir(parents = True, exist_ok = True)

        with open(self.target, 'w') as f:
            f.write(output)



import importlib.util

def load_blueprint(filename):
    spec = importlib.util.spec_from_file_location('blueprint', filename)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module



import pathlib
import asyncio
import time
import click

from .core import Context
from .util.load import load_blueprint
from .diagnostic.timeline import plot_timeline
from .diagnostic.graph import render_graph

@click.command()
@click.argument('targets', nargs = -1, type = click.Path(readable = False, path_type = pathlib.Path))
@click.option('-j', '--jobs', default = 1, help = 'Max parallel jobs.')
@click.option('--timeline', is_flag = True, help = 'Create a timeline plot after the build.')
@click.option('--graph', is_flag = True, help = 'Create a render of the dependency graph after the build.')
@click.option('--no-cache', is_flag = True, help = 'Don\'t use a cache file.')
def main(targets = None, jobs = None, timeline = False, graph = False, no_cache = False):
    main_start = time.monotonic()

    blueprint = pathlib.Path('blueprint.py')
    assert blueprint.exists()

    with Context(
        max_concurrent_tasks = jobs,
        cache_file = False if no_cache else None,
    ) as ctx:
        load_blueprint(blueprint)

        run_start = time.monotonic()

        tasks = []
        if targets:
            for target in targets:
                target_files = [file for file in ctx.files.values() if file.path.is_relative_to(target)]
                assert target_files, f'No targets matching {target}'
                tasks.extend(target_files)
        
        else:
            tasks = ctx.tasks.values()

        asyncio.run(ctx.run(tasks))

        if timeline:
            plot_timeline(ctx, main_start, run_start)

        if graph:
            render_graph(ctx)



import asyncio
import shlex

async def subprocess(cmd, *, stdout = None, stderr = None):
    print(shlex.join(str(e) for e in cmd))
    process = await asyncio.create_subprocess_exec(
        *(str(e) for e in cmd),
        stdout = stdout,
        stderr = stderr,
    )
    code = await process.wait()
    if code != 0:
        raise RuntimeError(f'Process returned {code}')



import pathlib
import asyncio

from .. import core
from ..util.subprocess import subprocess
from ..util.module_mapper import ModuleMapper

__all__ = ['Compile', 'Link']

class Env(core.Env):
    toolchain_prefix: str
    'GCC toolchain prefix'
    toolchain_suffix: str
    'GCC toolchain suffix'

    cflags: list[str]
    'C compiler flags'
    cxxflags: list[str]
    'C++ compiler flags'
    ldflags: list[str]
    'Linker flags'

    defines: list[str]
    'C/C++ preprocessor defines'
    include_path: list[str | pathlib.Path]
    'Include path'

    def __init__(self, *, cxx_modules = False, **kwargs):
        super().__init__(**kwargs)

        self.toolchain_prefix = ''
        self.toolchain_suffix = ''
        self.cflags = []
        self.cxxflags = []
        self.ldflags = []
        self.defines = []
        self.include_path = []

        if cxx_modules:
            self.module_mapper = ModuleMapper(self.build_dir / 'cmi')
            self.ctx.start_async(self.module_mapper.start())
        else:
            self.module_mapper = None

    def executable(self, output_file, source_files, **kwargs):
        return Link(self, output_file, [pathlib.Path(f) for f in source_files], **kwargs)

class ScanDeps(core.Task):
    env: Env

    def __new__(cls, env: Env, source_file):
        try:
            self = super().__new__(cls, env.ctx, ('scan_deps', env.build_dir, source_file))
        except core.TaskExists as e:
            if e.task.env == env:
                return e.task
            raise

        self.env = env
        self.source_file = source_file
        self.object_file = self.env.build_dir / 'objects' / source_file.with_suffix('.o')
        self.dep_file = self.env.build_dir / 'objects' / source_file.with_suffix('.d')
        self.add_input_files(source_file)
        self.add_output_files(self.dep_file)
        return self

    def input_metadata(self):
        return super().input_metadata() | {
            'toolchain_prefix': self.env.toolchain_prefix,
            'toolchain_suffix': self.env.toolchain_suffix,
            'flags': self.env.cflags if self.source_file.suffix == '.c' else self.env.cxxflags,
            'defines': self.env.defines,
            'include_path': self.env.include_path,
        }

    async def run(self):
        source_file = self.source_file
        object_file = self.object_file
        dep_file = self.dep_file
        cmi_dir = self.env.build_dir / 'cmi'

        # Ensure output directory exists.
        object_file.parent.mkdir(parents = True, exist_ok = True)

        if source_file.suffix == '.c':
            compiler = f'{self.env.toolchain_prefix}gcc{self.env.toolchain_suffix}'
            flags = self.env.cflags.copy()

        else:
            compiler = f'{self.env.toolchain_prefix}g++{self.env.toolchain_suffix}'
            flags = self.env.cxxflags.copy()
            if self.env.module_mapper is not None:
                flags.extend([
                    '-fmodules-ts',
                    self.env.module_mapper.gcc_arg,
                ])

        for define in self.env.defines:
            flags.extend(['-D', define])
        for path in self.env.include_path:
            flags.extend(['-I', path])

        await subprocess([
            compiler,
            *flags,
            '-MMD', '-E',
            source_file,
            '-MT', object_file,
            '-MF', dep_file,
        ], stdout = asyncio.subprocess.DEVNULL)

        depmap = {}

        with open(dep_file) as f:
            contents = f.read().replace('\\\n', ' ')
            for line in contents.split('\n'):
                if line.count(':') != 1:
                    continue
                targets, deps = line.split(':')
                targets = targets.split()
                deps = deps.replace('|', '').split()

                for t in targets:
                    if t not in depmap:
                        depmap[t] = []
                    depmap[t].extend(deps)

        file_deps = []
        module_deps = []
        module_gens = []

        for f in depmap[str(object_file)]:
            if f.endswith('.c++m'):
                f = self.env.build_dir / 'cmi' / f.replace('.c++m', '.gcm')

            file_deps.append(pathlib.Path(f))

        for dep in depmap[str(object_file)]:
            if dep.endswith('.c++m'):
                module_deps.append(dep[:-5])

        for dep in depmap.get('.PHONY', []):
            if dep.endswith('.c++m'):
                module_gens.append(dep[:-5])

        # Add included files as input files to trigger a rerun of this task if any changes in the future.
        self.add_input_files(*(f for f in file_deps if f.suffix != '.gcm'))

        return {
            'file_deps': file_deps,
            'module_deps': module_deps,
            'module_gens': module_gens,
        }

class Compile(core.Task):
    env: Env

    def __new__(cls, env, source_file):
        try:
            self = super().__new__(cls, env.ctx, ('compile', env.build_dir, source_file))
        except core.TaskExists as e:
            if e.task.env == env:
                return e.task
            raise

        self.env = env
        self.source_file = source_file
        self.object_file = self.env.build_dir / 'objects' / source_file.with_suffix('.o')
        self.scan_deps = ScanDeps(env, source_file)
        self.dependencies.append(self.scan_deps)
        #self.add_input_files(source_file) # This is added implicitly through scan_deps.
        self.add_output_files(self.object_file)
        return self

    def input_metadata(self):
        return super().input_metadata() | {
            'toolchain_prefix': self.env.toolchain_prefix,
            'toolchain_suffix': self.env.toolchain_suffix,
            'flags': self.env.cflags if self.source_file.suffix == '.c' else self.env.cxxflags,
            'defines': self.env.defines,
            'include_path': self.env.include_path,
        }

    def dynamic_deps(self):
        self.add_input_files(*self.scan_deps.result['file_deps'])

        return []

    async def run(self):
        source_file = self.source_file
        object_file = self.object_file
        cmi_dir = self.env.build_dir / 'cmi'

        # Ensure output directory exists.
        object_file.parent.mkdir(parents = True, exist_ok = True)

        if source_file.suffix == '.c':
            compiler = f'{self.env.toolchain_prefix}gcc{self.env.toolchain_suffix}'
            flags = self.env.cflags.copy()

        else:
            compiler = f'{self.env.toolchain_prefix}g++{self.env.toolchain_suffix}'
            flags = self.env.cxxflags.copy()
            if self.env.module_mapper is not None:
                flags.extend([
                    '-fmodules-ts',
                    self.env.module_mapper.gcc_arg,
                ])

        for define in self.env.defines:
            flags.extend(['-D', define])
        for path in self.env.include_path:
            flags.extend(['-I', path])

        await subprocess([
            compiler,
            *flags,
            '-c',
            source_file,
            '-o', object_file,
        ])

    async def post_run(self):
        # Report that modules are built.
        for m in self.scan_deps.result['module_gens']:
            self.add_output_files(self.env.build_dir / f'cmi/{m}.gcm')

class ModuleCheck(core.Task):
    env: Env

    def __new__(cls, env, target, source_files):
        self = super().__new__(cls, env.ctx, ('module_check', env.build_dir, target))
        self.env = env
        self.scan_deps_tasks = [ScanDeps(env, f) for f in source_files]
        self.dependencies.extend(self.scan_deps_tasks)
        return self

    async def run(self):
        generated_modules = set()

        for t in self.scan_deps_tasks:
            generated_modules.update(t.result['module_gens'])

        for t in self.scan_deps_tasks:
            for m in t.result['module_deps']:
                if m not in generated_modules:
                    raise Exception(f'Module {m} required by {t.source_file} does not exist.')

class Link(core.Task):
    env: Env

    def __new__(cls, env, target, source_files, ld_script = None):
        self = super().__new__(cls, env.ctx, ('link', env.build_dir, target))
        self.env = env
        self.target = target
        self.dependencies.append(ModuleCheck(env, target, source_files))
        self.source_files = source_files,
        self.object_tasks = [Compile(env, f) for f in source_files]
        self.ld_script = ld_script
        #self.dependencies.extend(self.object_tasks) # This is added implicitly through input_files
        self.elf_file = self.env.build_dir / self.target
        self.add_input_files(*(t.object_file for t in self.object_tasks))
        self.add_output_files(self.elf_file)
        if self.ld_script is not None:
            self.add_input_files(self.ld_script)
        return self

    def input_metadata(self):
        return super().input_metadata() | {
            'toolchain_prefix': self.env.toolchain_prefix,
            'toolchain_suffix': self.env.toolchain_suffix,
            'source_files': self.source_files,
            'ld_script': self.ld_script,
            'toolchain': self.env.toolchain_prefix,
            'flags': self.env.ldflags,
        }

    async def run(self):
        object_files_str = ' '.join(str(t.object_file) for t in self.object_tasks)
        elf_file = self.elf_file

        # Ensure output directory exists.
        elf_file.parent.mkdir(parents = True, exist_ok = True)

        ldflags = self.env.ldflags[:]

        if self.ld_script is not None:
            ldflags.extend(['-T', self.ld_script])

        await subprocess([
            f'{self.env.toolchain_prefix}g++{self.env.toolchain_suffix}',
            *ldflags,
            *(str(t.object_file) for t in self.object_tasks),
            '-o', elf_file,
        ])

        return elf_file



def _exclude_task(task):
    return task.id[0] in [
        'module_check',
        'scan_deps',
    ]

def render_graph(ctx):
    import graphviz

    dot = graphviz.Digraph(name = 'erect.deps')
    dot.attr(rankdir = 'LR')

    for path, file in ctx.files.items():
        # Exclude source files.
        if file.generator_task is None:
            continue

        # Exclude file generated by excluded tasks.
        if file.generator_task is not None and _exclude_task(file.generator_task):
            continue

        dot.node(str(path), color = 'blue')

        if file.generator_task is not None:
            dot.edge(file.generator_task.id.str, str(path))

    for id, task in ctx.tasks.items():
        if _exclude_task(task):
            continue

        dot.node(id.str, color = 'red')

        for dep in task.dependencies:
            if _exclude_task(dep):
                continue
            dot.edge(dep.id.str, id.str)

        for file in task._input_files:
            if file.generator_task is None:
                continue

            dot.edge(str(file.path), id.str)

    dot.render(view = True)



import pathlib
import asyncio
import hashlib
from dataclasses import dataclass

__all__ = ['Fingerprint', 'File']

# TODO: memoize?
def _hash_file(path):
    with open(path, 'rb') as f:
        return hashlib.file_digest(f, 'sha256').digest()

@dataclass
class Fingerprint:
    mtime_ns: int
    hash: bytes

    @classmethod
    def create(cls, path):
        assert path.exists()

        return cls(
            mtime_ns = path.stat().st_mtime_ns,
            hash = _hash_file(path),
        )

    def check(self, path):
        # Files that doesn't exist matches nothing.
        if not path.exists():
            return False
        
        # Assume file is unchanged if mtime matches.
        if path.stat().st_mtime_ns == self.mtime_ns:
            return True

        # Check if hash is matching.
        return _hash_file(path) == self.hash

class File:
    path: pathlib.Path

    def __new__(cls, ctx, path):
        path = pathlib.Path(path)
        if path in ctx.files:
            return ctx.files[path]
        ctx.files[path] = self = super().__new__(cls)
        self.ctx = ctx
        self.path = path

        dynamic = path.suffix == '.gcm' # TODO: refactor this up
        
        self._future = asyncio.Future() if dynamic else None

        self._generator_task = None
        return self

    @property
    def generator_task(self):
        return self._generator_task

    @generator_task.setter
    def generator_task(self, task):
        assert self._generator_task is None
        self._generator_task = task

        if self._future is not None:
            self._future.set_result(task)

    async def _run(self):
        if self._future is not None:
            await self._future

        if self.generator_task is not None:
            await self.generator_task._run()

        assert self.path.exists()

    def get_fingerprint(self):
        return Fingerprint.create(self.path)



def plot_timeline(ctx, main_start, run_start):
    from bokeh.models import BasicTickFormatter
    from bokeh.plotting import figure, show

    tasks = [
        (main_start, run_start, 'load_blueprint'),
    ]

    for task in ctx.tasks.values():
        if task._start is None or task._end is None:
            continue

        tasks.append((task._start, task._end, task.id.str))

    source = {'name': [], 'start': [], 'end': []}
    for start, end, name in reversed(sorted(tasks)):
        source['name'].append(name)
        source['start'].append(start - main_start)
        source['end'].append(end - main_start)

    p = figure(
        y_range = source['name'],
        toolbar_location=None,
        sizing_mode='stretch_width',
        height = 15 * len(tasks),
    )
    p.hbar(y='name', left='start', right='end', height=0.4, source=source)

    p.ygrid.grid_line_color = None
    p.xaxis.axis_label = 'Time (seconds)'
    p.xaxis.formatter = BasicTickFormatter(use_scientific=False)
    p.outline_line_color = None

    show(p)
